@@ -1,55 +1,17 @@
 # ***** BEGIN LICENSE BLOCK *****
-# Version: MPL 1.1/GPL 2.0
 #
-# The contents of this file are subject to the Mozilla Public License
-# Version 1.1 (the "License"); you may not use this file except in
-# compliance with the License. You may obtain a copy of the License at
-# http://www.mozilla.org/MPL/
-#
-# Software distributed under the License is distributed on an "AS IS"
-# basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
-# the License for the specific language governing rights and
-# limitations under the License.
-#
-# The Original Code is Pika.
-#
-# The Initial Developers of the Original Code are LShift Ltd, Cohesive
-# Financial Technologies LLC, and Rabbit Technologies Ltd.  Portions
-# created before 22-Nov-2008 00:00:00 GMT by LShift Ltd, Cohesive
-# Financial Technologies LLC, or Rabbit Technologies Ltd are Copyright
-# (C) 2007-2008 LShift Ltd, Cohesive Financial Technologies LLC, and
-# Rabbit Technologies Ltd.
-#
-# Portions created by LShift Ltd are Copyright (C) 2007-2009 LShift
-# Ltd. Portions created by Cohesive Financial Technologies LLC are
-# Copyright (C) 2007-2009 Cohesive Financial Technologies
-# LLC. Portions created by Rabbit Technologies Ltd are Copyright (C)
-# 2007-2009 Rabbit Technologies Ltd.
-#
-# Portions created by Tony Garnock-Jones are Copyright (C) 2009-2010
-# LShift Ltd and Tony Garnock-Jones.
-#
-# All Rights Reserved.
-#
-# Contributor(s): ______________________________________.
-#
-# Alternatively, the contents of this file may be used under the terms
-# of the GNU General Public License Version 2 or later (the "GPL"), in
-# which case the provisions of the GPL are applicable instead of those
-# above. If you wish to allow use of your version of this file only
-# under the terms of the GPL, and not to allow others to use your
-# version of this file under the terms of the MPL, indicate your
-# decision by deleting the provisions above and replace them with the
-# notice and other provisions required by the GPL. If you do not
-# delete the provisions above, a recipient may use your version of
-# this file under the terms of any one of the MPL or the GPL.
+# For copyright and licensing please refer to COPYING.
 #
 # ***** END LICENSE BLOCK *****
-import pika.callback as callback
-import pika.frames as frames
-import pika.log as log
-import pika.spec as spec
+
+import types
+
+import pika.frame as frame
 import pika.exceptions as exceptions
+import pika.log
+import pika.spec as spec
+
+MAX_CHANNELS = 32768
 
 
 class ChannelTransport(object):
@@ -65,17 +27,16 @@ class ChannelTransport(object):
     flow detection has been removed.
     """
 
-    @log.method_call
     def __init__(self, connection, channel_number):
 
         self.connection = connection
         self.channel_number = channel_number
 
         # Use the global callback handler
-        self.callbacks = callback.CallbackManager.instance()
+        self.callbacks = connection.callbacks
 
         # The frame-handler changes depending on the type of frame processed
-        self.frame_handler = ContentHandler()
+        self.frame_dispatcher = frame.Dispatcher(self.callbacks)
 
         # We need to block on synchronous commands, but do so asynchronously
         self.blocking = None
@@ -88,23 +49,20 @@ class ChannelTransport(object):
         self.callbacks.add(self.channel_number, spec.Channel.OpenOk,
                            self._on_open)
 
-    @log.method_call
     def _on_open(self, frame):
         """
         Called as a result of receiving a spec.Channel.OpenOk frame.
         """
         self.closed = False
 
-    @log.method_call
     def deliver(self, frame):
         """
         Deliver a frame to the frame handler. When it's gotten to a point that
         it has build our frame responses in a way we can use them, it will
         call our callback stack to figure out what to do.
         """
-        self.frame_handler.process(frame)
+        self.frame_dispatcher.process(frame)
 
-    @log.method_call
     def _ensure(self):
         """
         Make sure the transport is open
@@ -114,15 +72,13 @@ class ChannelTransport(object):
 
         return True
 
-    @log.method_call
     def _has_content(self, method):
         """
         Return a bool if it's a content method as defined by the spec
         """
         return spec.has_content(method.INDEX)
 
-    @log.method_call
-    def rpc(self, method, callback=None, acceptable_replies=[]):
+    def rpc(self, method, callback=None, acceptable_replies=None):
         """
         Shortcut wrapper to the Connection's rpc command using its callback
         stack, passing in our channel number
@@ -132,27 +88,36 @@ class ChannelTransport(object):
 
         # If we're blocking, add subsequent commands to our stack
         if self.blocking:
-            log.debug('%s: %s is blocking this channel',
-                          self.__class__.__name__, self.blocking)
+            pika.log.debug('%s: %s is blocking this channel',
+                            self.__class__.__name__, self.blocking)
 
             self._blocked.append([method, callback, acceptable_replies])
             return
 
+        # Validate we got None or a list of acceptable_replies
+        if acceptable_replies and not isinstance(acceptable_replies, list):
+            raise TypeError("acceptable_replies should be list or None")
+
+        # Validate the callback is a function or instancemethod
+        if callback and not isinstance(callback, types.FunctionType) and \
+           not isinstance(callback, types.MethodType):
+            raise TypeError("callback should be None, a function or method.")
+
         # If this is a synchronous method, block connections until we're done
         if method.synchronous:
-            log.debug('%s: %s turning on blocking',
-                          self.__class__.__name__, method.NAME)
+            pika.log.debug('%s: %s turning on blocking',
+                            self.__class__.__name__, method.NAME)
             self.blocking = method.NAME
 
-        for reply in acceptable_replies:
-            self.callbacks.add(self.channel_number, reply,
-                               self._on_synchronous_complete)
-            if callback:
-                self.callbacks.add(self.channel_number, reply, callback)
+        if acceptable_replies:
+            for reply in acceptable_replies:
+                self.callbacks.add(self.channel_number, reply,
+                                   self._on_synchronous_complete)
+                if callback:
+                    self.callbacks.add(self.channel_number, reply, callback)
 
         self.send_method(method)
 
-    @log.method_call
     def send_method(self, method, content=None):
         """
         Shortcut wrapper to send a method through our connection, passing in
@@ -160,7 +125,6 @@ class ChannelTransport(object):
         """
         self.connection._send_method(self.channel_number, method, content)
 
-    @log.method_call
     def _on_event_ok(self, frame):
         """
         Generic events that returned ok that may have internal callbacks.
@@ -169,7 +133,6 @@ class ChannelTransport(object):
         """
         pass
 
-    @log.method_call
     def _on_synchronous_complete(self, frame):
         """
         This is called when a synchronous command is completed. It will undo
@@ -196,7 +159,6 @@ class ChannelTransport(object):
 
 class Channel(spec.DriverMixin):
 
-    @log.method_call
     def __init__(self, connection, channel_number, on_open_callback=None,
                  transport=None):
         """
@@ -219,11 +181,8 @@ class Channel(spec.DriverMixin):
         # Channel Number
         self.channel_number = channel_number
 
-        # Assign our connection we communicate with
-        self.connection = connection
-
         # Get the callback manager
-        self.callbacks = callback.CallbackManager.instance()
+        self.callbacks = connection.callbacks
 
         # Our on_open_callback, special case
         self._on_open_callback = on_open_callback
@@ -244,14 +203,14 @@ class Channel(spec.DriverMixin):
                            '_on_basic_deliver',
                            self._on_basic_deliver,
                            False,
-                           ContentHandler)
+                           frame.Dispatcher)
 
         # Add a callback for Basic.Get
         self.callbacks.add(self.channel_number,
                            '_on_basic_get',
                            self._on_basic_get_ok,
                            False,
-                           ContentHandler)
+                           frame.Dispatcher)
 
         # Add a callback for Basic.GetEmpty
         self.callbacks.add(self.channel_number,
@@ -273,7 +232,6 @@ class Channel(spec.DriverMixin):
         # Open our channel
         self.transport.send_method(spec.Channel.Open())
 
-    @log.method_call
     def add_callback(self, callback, replies):
         """
         Pass in a callback handler and a list replies from the
@@ -283,7 +241,6 @@ class Channel(spec.DriverMixin):
         for reply in replies:
             self.callbacks.add(self.channel_number, reply, callback)
 
-    @log.method_call
     def add_on_close_callback(self, callback):
         """
         Pass a callback function that will be called when the channel is
@@ -291,7 +248,6 @@ class Channel(spec.DriverMixin):
         """
         self.callbacks.add(self.channel_number, '_on_channel_close', callback)
 
-    @log.method_call
     def add_on_return_callback(self, callback):
         """
         Pass a callback function that will be called when basic_publish as sent
@@ -302,7 +258,6 @@ class Channel(spec.DriverMixin):
         """
         self.callbacks.add(self.channel_number, '_on_basic_return', callback)
 
-    @log.method_call
     def close(self, code=0, text="Normal Shutdown", from_server=False):
         """
         Will invoke a clean shutdown of the channel with the AMQP Broker
@@ -323,7 +278,6 @@ class Channel(spec.DriverMixin):
         if not len(self._consumers) and not from_server:
             self._close()
 
-    @log.method_call
     def _close(self):
         """
         Internal close, is called when all the consumers are closed by both
@@ -333,7 +287,6 @@ class Channel(spec.DriverMixin):
                                                       self.closing[1],
                                                       0, 0))
 
-    @log.method_call
     def _on_remote_close(self, frame):
         """
         Handle the case where our channel has been closed for us
@@ -347,7 +300,6 @@ class Channel(spec.DriverMixin):
                                self, frame.method.reply_code,
                                frame.method.reply_text)
 
-    @log.method_call
     def _open(self, frame):
         """
         Called by our callback handler when we receive a Channel.OpenOk and
@@ -361,7 +313,6 @@ class Channel(spec.DriverMixin):
         if self._on_open_callback:
             self._on_open_callback(self)
 
-    @log.method_call
     def basic_cancel(self, consumer_tag, nowait=False, callback=None):
         """
         Pass in the consumer tag to cancel a basic_consume request with. The
@@ -384,16 +335,16 @@ class Channel(spec.DriverMixin):
                                              nowait=nowait),
                            self._on_cancel_ok, [spec.Basic.CancelOk])
 
-    @log.method_call
     def basic_consume(self, consumer_callback,
                       queue='', no_ack=False, exclusive=False,
                       consumer_tag=None):
         """
         Sends the AMQP command Basic.Consume to the broker and binds messages
         for the consumer_tag to the consumer callback. If you do not pass in
-        a consumer_tag, one will be automatically generated for you. For
-        more information on basic_consume, see:
+        a consumer_tag, one will be automatically generated for you. Returns
+        the consumer tag.
 
+        For more information on basic_consume, see:
         http://www.rabbitmq.com/amqp-0-9-1-reference.html#basic.consume
         """
         # If a consumer tag was not passed, create one
@@ -419,7 +370,9 @@ class Channel(spec.DriverMixin):
             del(self._consumers[consumer_tag])
             raise exceptions.ChannelClosed(e)
 
-    @log.method_call
+        # Return the consumer tag for the user reference
+        return consumer_tag
+
     def basic_publish(self, exchange, routing_key, body,
                       properties=None, mandatory=False, immediate=False):
         """
@@ -444,7 +397,6 @@ class Channel(spec.DriverMixin):
         """
         return self._consumers.keys()
 
-    @log.method_call
     def _on_basic_deliver(self, method_frame, header_frame, body):
         """
         Cope with reentrancy. If a particular consumer is still active when
@@ -483,12 +435,11 @@ class Channel(spec.DriverMixin):
             # Remove our pending messages
             del self._pending[consumer_tag]
         else:
-            # Append our _pending list with additional data
+            # Append the message to our pending list
             self._pending[consumer_tag].append((method_frame.method,
                                                 header_frame.properties,
                                                 body))
 
-    @log.method_call
     def _on_cancel_ok(self, frame):
         """
         Called in response to a frame from the Broker when we call Basic.Cancel
@@ -500,7 +451,6 @@ class Channel(spec.DriverMixin):
         if self.closing and not len(self._consumers):
             self._close()
 
-    @log.method_call
     def basic_get(self, callback, ticket=0, queue='', no_ack=False):
         """
         Get a single message from the AMQP broker. The callback method
@@ -518,7 +468,6 @@ class Channel(spec.DriverMixin):
                                                   queue=queue,
                                                   no_ack=no_ack))
 
-    @log.method_call
     def _on_basic_get_ok(self, method_frame, header_frame, body):
         if self._on_get_ok_callback:
             self._on_get_ok_callback(self,
@@ -527,17 +476,15 @@ class Channel(spec.DriverMixin):
                                      body)
             self._basic_get_ok_callback = None
         else:
-            log.error("%s._on_basic_get: No callback defined.",
-                          self.__class__.__name__)
+            pika.log.error("%s._on_basic_get: No callback defined.",
+                            self.__class__.__name__)
 
-    @log.method_call
     def _on_basic_get_empty(self, frame):
         """
         When we receive an empty reply do nothing but log it
         """
         pass
 
-    @log.method_call
     def flow(self, callback, active):
         """
         Turn Channel flow control off and on. Pass a callback to be notified
@@ -551,7 +498,6 @@ class Channel(spec.DriverMixin):
         self.transport.rpc(spec.Channel.Flow(active), self._on_channel_flow_ok,
                            [spec.Channel.FlowOk])
 
-    @log.method_call
     def _on_channel_flow_ok(self, frame):
         """
         Called in response to us asking the server to toggle on Channel.Flow
@@ -564,145 +510,33 @@ class Channel(spec.DriverMixin):
             self._on_flow_ok_callback(frame.method.active)
             self._on_flow_ok_callback = None
         else:
-            log.error("%s._on_flow_ok: No callback defined.",
-                      self.__class__.__name__)
+            pika.log.error("%s._on_flow_ok: No callback defined.",
+                           self.__class__.__name__)
 
-
-class ContentHandler(object):
-    """
-    This handles content frames which come in in synchronous order as follows:
-
-    1) Method Frame
-    2) Header Frame
-    3) Body Frame(s)
-
-    The way that content handler works is to assign the active frame type to
-    the self._handler variable. When we receive a header frame that is either
-    a Basic.Deliver, Basic.GetOk, or Basic.Return, we will assign the handler
-    to the ContentHandler._handle_header_frame. This will fire the next time
-    we are called and parse out the attributes required to receive the body
-    frames and assemble the content to be returned. We will then assign the
-    self._handler to ContentHandler._handle_body_frame.
-
-    _handle_body_frame has two methods inside of it, handle and finish.
-    handle will be invoked until the requirements set by the header frame have
-    been met at which point it will call the finish method. This calls
-    the callback manager with the method frame, header frame and assembled
-    body and then reset the self._handler to the _handle_method_frame method.
-    """
-
-    @log.method_call
-    def __init__(self):
-        # We start with Method frames always
-        self._handler = self._handle_method_frame
-
-    @log.method_call
-    def process(self, frame):
+    def confirm_delivery(self, callback=None, nowait=False):
         """
-        Invoked by the ChannelTransport object when passed frames that are not
-        setup in the rpc process and that don't have explicit reply types
-        defined. This includes Basic.Publish, Basic.GetOk and Basic.Return
-        """
-        self._handler(frame)
+        Turn on Confirm mode in the channel. Pass in a callback to be notified
+        by the Broker when a message has been confirmed as received (Basic.Ack
+        from the broker to the publisher).
 
-    @log.method_call
-    def _handle_method_frame(self, frame):
-        """
-        Receive a frame and process it, we should have content by the time we
-        reach this handler, set the next handler to be the header frame handler
-        """
-        # If we don't have FrameMethod something is wrong so throw an exception
-        if not isinstance(frame, frames.Method):
-            raise exceptions.UnexpectedFrameError(frame)
-
-        # If the frame is a content related frame go deal with the content
-        # By getting the content header frame
-        if spec.has_content(frame.method.INDEX):
-            self._handler = self._handle_header_frame(frame)
-
-        # We were passed a frame we don't know how to deal with
-        else:
-            raise NotImplementedError(frame.method.__class__)
-
-    @log.method_call
-    def _handle_header_frame(self, frame):
-        """
-        Receive a header frame and process that, setting the next handler
-        to the body frame handler
+        For more information see:
+            http://www.rabbitmq.com/extensions.html#confirms
         """
 
-        def handler(header_frame):
-            # Make sure it's a header frame
-            if not isinstance(header_frame, frames.Header):
-                raise exceptions.UnexpectedFrameError(header_frame)
+        # Add the Basic.Ack callback
+        if callback:
+            self.callbacks.add(self.channel_number,
+                               spec.Basic.Ack,
+                               callback,
+                               False)
 
-            # Call the handle body frame including our header frame
-            self._handle_body_frame(frame, header_frame)
+        # Send the RPC command
+        self.transport.rpc(spec.Confirm.Select(nowait),
+                           self._on_confirm_select_ok,
+                           [spec.Confirm.SelectOk])
 
-        return handler
-
-    @log.method_call
-    def _handle_body_frame(self, method_frame, header_frame):
+    def _on_confirm_select_ok(self, frame):
         """
-        Receive body frames. We'll keep receiving them in handler until we've
-        received the body size specified in the header frame. When done
-        call our finish function which will call our transports callbacks
+        Called when the broker sends a Confirm.SelectOk frame
         """
-        seen_so_far = [0]
-        body_fragments = list()
-
-        @log.method_call
-        def handler(body_frame):
-            # Make sure it's a body frame
-            if not isinstance(body_frame, frames.Body):
-                raise exceptions.UnexpectedFrameError(body_frame)
-
-            # Increment our counter so we know when we've had enough
-            seen_so_far[0] += len(body_frame.fragment)
-
-            # Append the fragment to our list
-            body_fragments.append(body_frame.fragment)
-
-            # Did we get enough bytes? If so finish
-            if seen_so_far[0] == header_frame.body_size:
-                finish()
-
-            # Did we get too many bytes?
-            elif seen_so_far[0] > header_frame.body_size:
-                error = 'Received %i and only expected %i' % \
-                        (seen_so_far[0], header_frame.body_size)
-                raise exceptions.BodyTooLongError(error)
-
-        @log.method_call
-        def finish():
-            # We're done so set our handler back to the method frame
-            self._handler = self._handle_method_frame
-
-            # Get our method name
-            method = method_frame.method.__class__.__name__
-            if method == 'Deliver':
-                key = '_on_basic_deliver'
-            elif method == 'GetOk':
-                key = '_on_basic_get'
-            elif method == 'Return':
-                key = '_on_basic_return'
-            else:
-                raise Exception('Unimplemented Content Return Key')
-
-            # Get our callback manager instance
-            callbacks = callback.CallbackManager.instance()
-
-            # Check for a processing callback for our method name
-            callbacks.process(method_frame.channel_number,  # Prefix
-                              key,                          # Key
-                              self,                         # Caller
-                              method_frame,                 # Arg 1
-                              header_frame,                 # Arg 2
-                              ''.join(body_fragments))      # Arg 3
-
-        # If we don't have a header frame body size, finish. Otherwise keep
-        # going and keep our handler function as the frame handler
-        if not header_frame.body_size:
-            finish()
-        else:
-            self._handler = handler
+        pika.log.info("Confirm.SelectOk Received")
